==> heteroscalapack-1.0.2-BETA.tar.gz: heteroscalapack-1.0.2-BETA/SRC/hscal_pzgesv_ctxt.c <==
/*
* Ravi Reddy
* Version 1.0
* School of Computer Science and Informatics, 
* UCD, 
* Dublin,
* November 9, 2007
*/

/*
*  Include files
*/
#include "../include/hscalapack.h"

int hscal_pzgesv_ctxt(
   int    * N,
   int    * NRHS,
   int    * IA,
   int    * JA,
   int    * DESCA,
   int    * IB,
   int    * JB,
   int    * DESCB,
   int    * ICTXT
)
{
/*  -- ScaLAPACK routine (version 1.7) -- */
/*     University of Tennessee, Knoxville, Oak Ridge National Laboratory, */
/*     and University of California, Berkeley. */
/*     Jan 30, 2006 */

/*     .. Scalar Arguments .. */
/*     .. */
/*     .. Array Arguments .. */
/*     .. */

/*  Purpose */
/*  ======= */

/*  PZGESV computes the solution to a complex system of linear equations */

/*                        sub( A ) * X = sub( B ), */

/*  where sub( A ) = A(IA:IA+N-1,JA:JA+N-1) is an N-by-N distributed */
/*  matrix and X and sub( B ) = B(IB:IB+N-1,JB:JB+NRHS-1) are N-by-NRHS */
/*  distributed matrices. */

/*  The LU decomposition with partial pivoting and row interchanges is */
/*  used to factor sub( A ) as sub( A ) = P * L * U, where P is a permu- */
/*  tation matrix, L is unit lower triangular, and U is upper triangular. */
/*  L and U are stored in sub( A ). The factored form of sub( A ) is then */
/*  used to solve the system of equations sub( A ) * X = sub( B ). */

/*  Notes */
/*  ===== */

/*  Each global data object is described by an associated description */
/*  vector.  This vector stores the information required to establish */
/*  the mapping between an object element and its corresponding process */
/*  and memory location. */

/*  Let A be a generic term for any 2D block cyclicly distributed array. */
/*  Such a global array has an associated description vector DESCA. */
/*  In the following comments, the character _ should be read as */
/*  "of the global array". */

/*  NOTATION        STORED IN      EXPLANATION */
/*  --------------- -------------- -------------------------------------- */
/*  DTYPE_A(global) DESCA( DTYPE_ )The descriptor type.  In this case, */
/*                                 DTYPE_A = 1. */
/*  CTXT_A (global) DESCA( CTXT_ ) The BLACS context handle, indicating */
/*                                 the BLACS process grid A is distribu- */
/*                                 ted over. The context itself is glo- */
/*                                 bal, but the handle (the integer */
/*                                 value) may vary. */
/*  M_A    (global) DESCA( M_ )    The number of rows in the global */
/*                                 array A. */
/*  N_A    (global) DESCA( N_ )    The number of columns in the global */
/*                                 array A. */
/*  MB_A   (global) DESCA( MB_ )   The blocking factor used to distribute */
/*                                 the rows of the array. */
/*  NB_A   (global) DESCA( NB_ )   The blocking factor used to distribute */
/*                                 the columns of the array. */
/*  RSRC_A (global) DESCA( RSRC_ ) The process row over which the first */
/*                                 row of the array A is distributed. */
/*  CSRC_A (global) DESCA( CSRC_ ) The process column over which the */
/*                                 first column of the array A is */
/*                                 distributed. */
/*  LLD_A  (local)  DESCA( LLD_ )  The leading dimension of the local */
/*                                 array.  LLD_A >= MAX(1,LOCr(M_A)). */

/*  Let K be the number of rows or columns of a distributed matrix, */
/*  and assume that its process grid has dimension p x q. */
/*  LOCr( K ) denotes the number of elements of K that a process */
/*  would receive if K were distributed over the p processes of its */
/*  process column. */
/*  Similarly, LOCc( K ) denotes the number of elements of K that a */
/*  process would receive if K were distributed over the q processes of */
/*  its process row. */
/*  The values of LOCr() and LOCc() may be determined via a call to the */
/*  ScaLAPACK tool function, NUMROC: */
/*          LOCr( M ) = NUMROC( M, MB_A, MYROW, RSRC_A, NPROW ), */
/*          LOCc( N ) = NUMROC( N, NB_A, MYCOL, CSRC_A, NPCOL ). */
/*  An upper bound for these quantities may be computed by: */
/*          LOCr( M ) <= ceil( ceil(M/MB_A)/NPROW )*MB_A */
/*          LOCc( N ) <= ceil( ceil(N/NB_A)/NPCOL )*NB_A */

/*  This routine requires square block decomposition ( MB_A = NB_A ). */

/*  Arguments */
/*  ========= */

/*  N       (global input) INTEGER */
/*          The number of rows and columns to be operated on, i.e. the */
/*          order of the distributed submatrix sub( A ). N >= 0. */

/*  NRHS    (global input) INTEGER */
/*          The number of right hand sides, i.e., the number of columns */
/*          of the distributed submatrix sub( B ). NRHS >= 0. */

/*  IA      (global input) INTEGER */
/*          The row index in the global array A indicating the first */
/*          row of sub( A ). */

/*  JA      (global input) INTEGER */
/*          The column index in the global array A indicating the */
/*          first column of sub( A ). */

/*  DESCA   (global and local input) INTEGER array of dimension DLEN_. */
/*          The array descriptor for the distributed matrix A. */

/*  IB      (global input) INTEGER */
/*          The row index in the global array B indicating the first */
/*          row of sub( B ). */

/*  JB      (global input) INTEGER */
/*          The column index in the global array B indicating the */
/*          first column of sub( B ). */

/*  DESCB   (global and local input) INTEGER array of dimension DLEN_. */
/*          The array descriptor for the distributed matrix B. */

/*  ===================================================================== */
/*
*  .. Local Scalars ..
*/
    int            rc, p, q;
    HMPI_Group     *gid;
    MPI_Comm       *pzgesvcomm;
    HSCALCONTEXT   *hctxt; 
    double         algotime;
   
/*
* Initialize the context ID
*/
    *ICTXT = HSCAL_CTXT_UNDEFINED;

/*
* Dynamically refresh the speeds of the processors
*/
    rc = hscal_pzgesv_recon(
              N,
              NRHS,
              IA,
              JA,
              DESCA,
              IB,
              JB,
              DESCB
    );

    if (rc != HSCAL_SUCCESS)
    {
       return HSCAL_ERR_RECON;
    }

/*
* Create a HeteroMPI group with the optimal grid arrangement 
* of processes
*/
    gid  = (HMPI_Group*)malloc(sizeof(HMPI_Group));

    if (gid == NULL)
    {
       return HSCAL_ERR_LOW_MEMORY;
    }

    rc = hscal_pzgesv_group_heuristic_auto_create(
              N,
              NRHS,
              IA,
              JA,
              DESCA,
              IB,
              JB,
              DESCB,
              gid,
              hscal_model_pzgesv,
              hscal_heuristic_fn_nonlinear
    );

    if (rc != HSCAL_SUCCESS)
    {
       return HSCAL_ERR_GROUP_CREATION;
    }

    if (HMPI_Is_free())
    {
       free(gid);

       *ICTXT = HSCAL_CTXT_UNDEFINED;

       return HSCAL_SUCCESS;
    }

    pzgesvcomm = (MPI_Comm*)HMPI_Get_comm(gid);

    if (pzgesvcomm == NULL)
    {
       return HSCAL_ERR_NULL_COMM; 
    } 

/* 
* Broadcast the estimated execution time 
*/ 
    algotime = HMPI_Group_timeof(gid);

/*
* Broadcast the optimal grid arrangement
*/
    if (HMPI_Is_parent(gid))
    {
       int nd;
       int *d;

       HMPI_Group_topology(
           gid,
           &nd,
           &d
       );

       p = d[0];

       free(d);

       rc = MPI_Bcast(
               &p,
               1,
               MPI_INT,
               0,
               *pzgesvcomm
       );
    }
    else
    {
       rc = MPI_Bcast(
               &p,
               1,
               MPI_INT,
               0,
               *pzgesvcomm
       );
    }

    if (rc != MPI_SUCCESS)
    {
       return HSCAL_ERR_BCAST_OPTIMAL_GRID;
    }

    if (HMPI_Is_parent(gid))
    {
       int nd;
       int *d;

       HMPI_Group_topology(
           gid,
           &nd,
           &d
       );

       q = d[1];

       free(d);

       rc = MPI_Bcast(
               &q,
               1,
               MPI_INT,
               0,
               *pzgesvcomm
       );
    }
    else
    {
       rc = MPI_Bcast(
               &q,
               1,
               MPI_INT,
               0,
               *pzgesvcomm
       );
    }

    if (rc != MPI_SUCCESS)
    {
       return HSCAL_ERR_BCAST_OPTIMAL_GRID;
    }

    if (HMPI_Is_parent(gid))
    {
       HSCAL_DEBUG(
            __FILE__, __FUNCTION__, __LINE__,
            "Optimal number of process rows =%d, Optimal number of process columns =%d, " 
            "Estimated execution time=%f ",
            p, q, algotime
       );       
    }

/*
* Translate communicator to a BLACS handle
*/
    *ICTXT = Csys2blacs_handle(*pzgesvcomm);

/*
* Form BLACS context
*/
    Cblacs_gridinit(ICTXT, "r", p, q);

/*
* Store the context to gid mapping
*/
    hctxt  = (HSCALCONTEXT*)malloc(sizeof(HSCALCONTEXT)); 
 
    if (hctxt == NULL) 
    { 
       return HSCAL_ERR_LOW_MEMORY; 
    } 
 
    hctxt->time = algotime; 
    hctxt->optp = p; 
    hctxt->optq = q; 
    hctxt->gid  = gid; 
 
    rc = hscal_store_gid_to_ctxt_mapping(ICTXT, hctxt); 

    return rc;
}

==> heteroscalapack-1.0.7-BETA.tar.gz: heteroscalapack-1.0.7-BETA/SRC/hscal_pzgesv_ctxt.c <==
/*
* Ravi Reddy
* Version 1.0
* School of Computer Science and Informatics, 
* UCD, 
* Dublin,
* November 9, 2007
*/

/*
*  Include files
*/
#include "../include/hscalapack.h"

int hscal_pzgesv_ctxt(
   int    * N,
   int    * NRHS,
   int    * IA,
   int    * JA,
   int    * DESCA,
   int    * IB,
   int    * JB,
   int    * DESCB,
   int    * ICTXT
)
{
/*  -- ScaLAPACK routine (version 1.7) -- */
/*     University of Tennessee, Knoxville, Oak Ridge National Laboratory, */
/*     and University of California, Berkeley. */
/*     Jan 30, 2006 */

/*     .. Scalar Arguments .. */
/*     .. */
/*     .. Array Arguments .. */
/*     .. */

/*  Purpose */
/*  ======= */

/*  PZGESV computes the solution to a complex system of linear equations */

/*                        sub( A ) * X = sub( B ), */

/*  where sub( A ) = A(IA:IA+N-1,JA:JA+N-1) is an N-by-N distributed */
/*  matrix and X and sub( B ) = B(IB:IB+N-1,JB:JB+NRHS-1) are N-by-NRHS */
/*  distributed matrices. */

/*  The LU decomposition with partial pivoting and row interchanges is */
/*  used to factor sub( A ) as sub( A ) = P * L * U, where P is a permu- */
/*  tation matrix, L is unit lower triangular, and U is upper triangular. */
/*  L and U are stored in sub( A ). The factored form of sub( A ) is then */
/*  used to solve the system of equations sub( A ) * X = sub( B ). */

/*  Notes */
/*  ===== */

/*  Each global data object is described by an associated description */
/*  vector.  This vector stores the information required to establish */
/*  the mapping between an object element and its corresponding process */
/*  and memory location. */

/*  Let A be a generic term for any 2D block cyclicly distributed array. */
/*  Such a global array has an associated description vector DESCA. */
/*  In the following comments, the character _ should be read as */
/*  "of the global array". */

/*  NOTATION        STORED IN      EXPLANATION */
/*  --------------- -------------- -------------------------------------- */
/*  DTYPE_A(global) DESCA( DTYPE_ )The descriptor type.  In this case, */
/*                                 DTYPE_A = 1. */
/*  CTXT_A (global) DESCA( CTXT_ ) The BLACS context handle, indicating */
/*                                 the BLACS process grid A is distribu- */
/*                                 ted over. The context itself is glo- */
/*                                 bal, but the handle (the integer */
/*                                 value) may vary. */
/*  M_A    (global) DESCA( M_ )    The number of rows in the global */
/*                                 array A. */
/*  N_A    (global) DESCA( N_ )    The number of columns in the global */
/*                                 array A. */
/*  MB_A   (global) DESCA( MB_ )   The blocking factor used to distribute */
/*                                 the rows of the array. */
/*  NB_A   (global) DESCA( NB_ )   The blocking factor used to distribute */
/*                                 the columns of the array. */
/*  RSRC_A (global) DESCA( RSRC_ ) The process row over which the first */
/*                                 row of the array A is distributed. */
/*  CSRC_A (global) DESCA( CSRC_ ) The process column over which the */
/*                                 first column of the array A is */
/*                                 distributed. */
/*  LLD_A  (local)  DESCA( LLD_ )  The leading dimension of the local */
/*                                 array.  LLD_A >= MAX(1,LOCr(M_A)). */

/*  Let K be the number of rows or columns of a distributed matrix, */
/*  and assume that its process grid has dimension p x q. */
/*  LOCr( K ) denotes the number of elements of K that a process */
/*  would receive if K were distributed over the p processes of its */
/*  process column. */
/*  Similarly, LOCc( K ) denotes the number of elements of K that a */
/*  process would receive if K were distributed over the q processes of */
/*  its process row. */
/*  The values of LOCr() and LOCc() may be determined via a call to the */
/*  ScaLAPACK tool function, NUMROC: */
/*          LOCr( M ) = NUMROC( M, MB_A, MYROW, RSRC_A, NPROW ), */
/*          LOCc( N ) = NUMROC( N, NB_A, MYCOL, CSRC_A, NPCOL ). */
/*  An upper bound for these quantities may be computed by: */
/*          LOCr( M ) <= ceil( ceil(M/MB_A)/NPROW )*MB_A */
/*          LOCc( N ) <= ceil( ceil(N/NB_A)/NPCOL )*NB_A */

/*  This routine requires square block decomposition ( MB_A = NB_A ). */

/*  Arguments */
/*  ========= */

/*  N       (global input) INTEGER */
/*          The number of rows and columns to be operated on, i.e. the */
/*          order of the distributed submatrix sub( A ). N >= 0. */

/*  NRHS    (global input) INTEGER */
/*          The number of right hand sides, i.e., the number of columns */
/*          of the distributed submatrix sub( B ). NRHS >= 0. */

/*  IA      (global input) INTEGER */
/*          The row index in the global array A indicating the first */
/*          row of sub( A ). */

/*  JA      (global input) INTEGER */
/*          The column index in the global array A indicating the */
/*          first column of sub( A ). */

/*  DESCA   (global and local input) INTEGER array of dimension DLEN_. */
/*          The array descriptor for the distributed matrix A. */

/*  IB      (global input) INTEGER */
/*          The row index in the global array B indicating the first */
/*          row of sub( B ). */

/*  JB      (global input) INTEGER */
/*          The column index in the global array B indicating the */
/*          first column of sub( B ). */

/*  DESCB   (global and local input) INTEGER array of dimension DLEN_. */
/*          The array descriptor for the distributed matrix B. */

/*  ===================================================================== */
/*
*  .. Local Scalars ..
*/
    int            rc, p, q;
    HMPI_Group     *gid;
    MPI_Comm       *pzgesvcomm;
    HSCALCONTEXT   *hctxt; 
    double         algotime;
   
/*
* Initialize the context ID
*/
    *ICTXT = HSCAL_CTXT_UNDEFINED;

/*
* Dynamically refresh the speeds of the processors
*/
    rc = hscal_pzgesv_recon(
              N,
              NRHS,
              IA,
              JA,
              DESCA,
              IB,
              JB,
              DESCB
    );

    if (rc != HSCAL_SUCCESS)
    {
       return HSCAL_ERR_RECON;
    }

/*
* Create a HeteroMPI group with the optimal grid arrangement 
* of processes
*/
    gid  = (HMPI_Group*)malloc(sizeof(HMPI_Group));

    if (gid == NULL)
    {
       return HSCAL_ERR_LOW_MEMORY;
    }

    rc = hscal_pzgesv_group_heuristic_pauto_create(
              N,
              NRHS,
              IA,
              JA,
              DESCA,
              IB,
              JB,
              DESCB,
              gid,
              hscal_model_pzgesv_optimized,
              hscal_heuristic_fn_nonlinear
    );

    if (rc != HSCAL_SUCCESS)
    {
       return HSCAL_ERR_GROUP_CREATION;
    }

    if (HMPI_Is_free())
    {
       free(gid);

       *ICTXT = HSCAL_CTXT_UNDEFINED;

       return HSCAL_SUCCESS;
    }

    pzgesvcomm = (MPI_Comm*)HMPI_Get_comm(gid);

    if (pzgesvcomm == NULL)
    {
       return HSCAL_ERR_NULL_COMM; 
    } 

/* 
* Broadcast the estimated execution time 
*/ 
    algotime = HMPI_Group_timeof(gid);

/*
* Broadcast the optimal grid arrangement
*/
    if (HMPI_Is_parent(gid))
    {
       int nd;
       int *d;

       HMPI_Group_topology(
           gid,
           &nd,
           &d
       );

       p = d[0];

       free(d);

       rc = MPI_Bcast(
               &p,
               1,
               MPI_INT,
               0,
               *pzgesvcomm
       );
    }
    else
    {
       rc = MPI_Bcast(
               &p,
               1,
               MPI_INT,
               0,
               *pzgesvcomm
       );
    }

    if (rc != MPI_SUCCESS)
    {
       return HSCAL_ERR_BCAST_OPTIMAL_GRID;
    }

    if (HMPI_Is_parent(gid))
    {
       int nd;
       int *d;

       HMPI_Group_topology(
           gid,
           &nd,
           &d
       );

       q = d[1];

       free(d);

       rc = MPI_Bcast(
               &q,
               1,
               MPI_INT,
               0,
               *pzgesvcomm
       );
    }
    else
    {
       rc = MPI_Bcast(
               &q,
               1,
               MPI_INT,
               0,
               *pzgesvcomm
       );
    }

    if (rc != MPI_SUCCESS)
    {
       return HSCAL_ERR_BCAST_OPTIMAL_GRID;
    }

    if (HMPI_Is_parent(gid))
    {
       HSCAL_DEBUG(
            __FILE__, __FUNCTION__, __LINE__,
            "Optimal number of process rows =%d, Optimal number of process columns =%d, " 
            "Estimated execution time=%f ",
            p, q, algotime
       );       
    }

/*
* Translate communicator to a BLACS handle
*/
    *ICTXT = Csys2blacs_handle(*pzgesvcomm);

/*
* Form BLACS context
*/
    Cblacs_gridinit(ICTXT, "r", p, q);

/*
* Store the context to gid mapping
*/
    hctxt  = (HSCALCONTEXT*)malloc(sizeof(HSCALCONTEXT)); 
 
    if (hctxt == NULL) 
    { 
       return HSCAL_ERR_LOW_MEMORY; 
    } 
 
    hctxt->time = algotime; 
    hctxt->optp = p; 
    hctxt->optq = q; 
    hctxt->gid  = gid; 
 
    rc = hscal_store_gid_to_ctxt_mapping(ICTXT, hctxt); 

    return rc;
}
